import 'dart:convert';
import 'dart:typed_data';

import 'type_adapter.dart';

extension _ByteDataReader on BinaryReader {
  ByteData readByteData(int length) =>
      this.readBytes(length).buffer.asByteData();
}

extension _ByteDataWriter on BinaryWriter {
  void writeByteData(ByteData bytes) =>
      this.writeBytes(bytes.buffer.asUint8List());
}

// Several adapters for low-level type data. They are not intended to be
// registered, thus the type id is -1.

TypeAdapter<int> adapterForByte() => TypeAdapter<int>(
      typeId: -1,
      write: (writer, byte) =>
          writer.writeByteData(ByteData(1)..setUint8(0, byte)),
      read: (reader) => reader.readByteData(1).getUint8(0),
    );

TypeAdapter<int> adapterForWord() => TypeAdapter<int>(
      typeId: -1,
      write: (writer, word) =>
          writer.writeByteData(ByteData(2)..setUint16(0, word)),
      read: (reader) => reader.readByteData(2).getUint16(0),
    );

TypeAdapter<int> adapterForInt32() => TypeAdapter<int>(
      typeId: -1,
      write: (writer, int32) =>
          writer.writeByteData(ByteData(2)..setUint16(0, int32)),
      read: (reader) => reader.readByteData(2).getUint16(0),
    );

TypeAdapter<int> adapterForUint32() => TypeAdapter<int>(
      typeId: -1,
      write: (writer, uint32) =>
          writer.writeByteData(ByteData(2)..setUint16(0, uint32)),
      read: (reader) => reader.readByteData(2).getUint16(0),
    );

/// Adapter for a type id.
TypeAdapter<int> adapterForTypeId() => adapterForWord();

/// Adapter for a nullable type. Saves a boolean to indicate whether there's a
/// value or not.
TypeAdapter<T> adapterForNullable<T>(TypeAdapter<T> adapterForT) =>
    TypeAdapter<T>(
      typeId: 0,
      write: (writer, value) {
        if (value == null) {
          writer.write(adapterForBool(), false);
        } else {
          writer..write(adapterForBool(), true)..write(adapterForT, value);
        }
      },
      read: (reader) {
        final isNull = reader.read(adapterForBool());
        return isNull ? null : reader.read(adapterForT);
      },
    );

/// Adapter for dynamic values. To enable choosing the right adapter when
/// decoding, the type is saved alongside the value.
TypeAdapter<T> adapterForDynamic<T>() => TypeAdapter<T>(
      typeId: 1,
      write: (writer, obj) {
        if (obj == null) {
          writer.write(adapterForTypeId(), 0);
        } else {
          final adapter = writer.registry.findAdapterByType<T>();
          writer..write(adapterForTypeId(), adapter)..write(adapter, obj);
        }
      },
      read: (reader) {
        final typeId = reader.read(adapterForTypeId());
        if (typeId == 0) return null;
        final adapter = reader.registry.findAdapterById(typeId);
        return reader.read(adapter);
      },
    );

/// Adapter for [double]s.
TypeAdapter<double> adapterForDouble() => TypeAdapter<double>(
      typeId: 2,
      write: (writer, value) => writer
          .writeByteData(ByteData(8)..setFloat64(0, value, Endian.little)),
      read: (reader) => reader.readByteData(8).getFloat64(0, Endian.little),
    );

/// Adapter for [int]s.
TypeAdapter<int> adapterForInt() => TypeAdapter<int>(
      typeId: 3,
      write: (writer, value) =>
          writer.write(adapterForDouble(), value.toDouble()),
      read: (reader) => reader.read(adapterForDouble()).toInt(),
    );

/// Adapter for [bool]s.
TypeAdapter<bool> adapterForBool() => TypeAdapter<bool>(
      typeId: 4,
      write: (writer, obj) => writer.writeBytes([obj ? 1 : 0]),
      read: (reader) {
        final byte = reader.readBytes(1).single;
        assert(byte == 0 || byte == 1);
        return byte != 0;
      },
    );

/// Adapter for [Uint8List].
TypeAdapter<Uint8List> adapterForUint8List() => TypeAdapter<Uint8List>(
      typeId: 5,
      write: (writer, bytes) {
        writer
          ..write(adapterForUint32(), bytes.length)
          ..writeBytes(bytes);
      },
      read: (reader) {
        final length = reader.read(adapterForUint32());
        return reader.readBytes(length);
      },
    );

/// Adapter for ascii strings.
TypeAdapter<String> adapterForAsciiString() => TypeAdapter<String>(
      typeId: -1,
      write: (writer, string) {
        for (final codeUnit in string.codeUnits) {
          assert((codeUnit & ~0x7f) != 0, 'String contains non-ASCII chars.');
        }
        writer.write(adapterForUint8List(), string.codeUnits);
      },
      read: (reader) =>
          String.fromCharCodes(reader.read(adapterForUint8List())),
    );

/// Adapter for regular [String]s.
TypeAdapter<String> adapterForString() => TypeAdapter<String>(
      typeId: 6,
      write: (writer, string) =>
          writer.write(adapterForUint8List(), utf8.encode(string)),
      read: (reader) => utf8.decode(reader.read(adapterForUint8List())),
    );

/// Helper for primitive adapters.
TypeAdapter<T> _getAdapterForPrimitive<T>() {
  switch (T) {
    case String:
      return adapterForString() as TypeAdapter<T>;
    case int:
      return adapterForInt() as TypeAdapter<T>;
    case double:
      return adapterForDouble() as TypeAdapter<T>;
    case bool:
      return adapterForBool() as TypeAdapter<T>;
  }
  return null;
}

/// Adapter for [List]s. We need to save the type of every list item with it,
/// because a list of type T may contain elements that are instances of
/// subclasses of T.
/// Because [String]s, [int]s, [bool]s, [double]s can't be subclassed, the
/// type doesn't have to be saved for every item.
TypeAdapter<List<T>> adapterForList<T>(TypeAdapter<T> adapterForT) =>
    TypeAdapter<List<T>>(
      typeId: 7,
      write: (writer, list) {
        writer.write(adapterForUint32(), list.length);

        for (final item in list) {
          writer.write(adapterForT, item);
        }
      },
      read: (reader) {
        final length = reader.read(adapterForInt());
        return [
          for (var i = 0; i < length; i++) reader.read(adapterForT),
        ];
      },
    );

/// Adapter for [Set]s.
TypeAdapter<Set<T>> adapterForSet<T>(TypeAdapter<T> adapterForT) =>
    TypeAdapter<Set<T>>(
      typeId: 8,
      write: (writer, value) =>
          writer.write(adapterForList<T>(adapterForT), value.toList()),
      read: (reader) => reader.read(adapterForList<T>(adapterForT)).toSet(),
    );

/// Adapter for [Map]s.
TypeAdapter<Map<K, V>> adapterForMap<K, V>(
        TypeAdapter<K> adapterForK, TypeAdapter<V> adapterForV) =>
    TypeAdapter<Map<K, V>>(
      typeId: 9,
      write: (writer, map) {
        writer.write(adapterForUint32(), map.length);
        for (var entry in map.entries) {
          writer
            ..write(adapterForK, entry.key)
            ..write(adapterForV, entry.value);
        }
      },
      read: (reader) {
        final length = reader.read(adapterForUint32());
        return {
          for (var i = 0; i < length; i++)
            reader.read(adapterForK): reader.read(adapterForV),
        };
      },
    );

/// Adapter for generated number of fields.
TypeAdapter<int> adapterForNumberOfFields() => adapterForUint32();

/// Adapter for id of a field.
TypeAdapter<int> adapterForFieldId() => adapterForWord();
